#ifndef _BUCKETERS_HPP_
#define _BUCKETERS_HPP_

#include <iostream>
#include <opencv2/core/core.hpp>
#include "histodet/hsv.hpp"

namespace histodet {

enum BucketerType { btGray = 0, btHue, btHueBruteforce, btHueLowsat, btHueValue, btRGB, btHueHighsat };

/**
 This class is an interface for converting a pixel into a bucket in an histogram
 (and convert a source image into a histogram bucket index image).
 Then we just need to provide different implementations for this class to have
 different ways of building histograms for gray or color images (intensity,
 r*g*b, hue, h/s/v, h*v...)
*/
class BucketerAbstract
{
 protected:
	int hist_size;     ///< number of buckets of the histogram
	cv::Mat bucketImg; ///< histogram bucket index image
	bool rgb_order;    ///< if color image channels are in RGB order instead of BGR
 public:
	/**
	 * @return the enum value describing the bucketer type
	 */
	virtual BucketerType type() = 0;
	/**
	 * Constructor
	 * @param hist_size the histogram size (number of buckets)
	 */
	BucketerAbstract(int hist_size): hist_size(hist_size), rgb_order(false) {}
	/**
	 * Desctructor
	 */
	virtual ~BucketerAbstract() {}
	/**
	 * @param rgb_order if img is color, set this to true if channels are in the order RGB, and false if BGR
	 */
	void setRgbOrder(bool rgb_order)
		{ this->rgb_order = rgb_order; }
	/**
	 * @return the histogram size
	 */
	int getHistSize() { return hist_size; }
	/**
	 * Convert the image into a histogram bucket index image
	 * @param img the input normal image
	 * @return the equivalent bucket index image
	 */
	virtual cv::Mat& convertImage(cv::Mat const &img) = 0;

	virtual double param(int i) { return 0.0; }
	
	/**
		* Create a bucketer from a learned histogram file
		* @param hist_file the learned histogram description file
		* @param hist_size returns the histogram size
		* @param hist_nbits returns the histogram nbits log2(size)
		* @return the bucketer
		*/
	static BucketerAbstract* createBucketer(std::string hist_file, int & hist_size, int & hist_nbits);
};


/**
 Generic bucketer based on LookUpTable for color images.
 Only needs to inherit to implement constructors to build the LUT, chosing
 the template type according to histogram size.
*/
template <typename T, bool color, bool mask = false>
class BucketerLut: public BucketerAbstract
{
 protected:
	int lut_size;
	int lut_tobin;
	int lut_size2;
	T *lut;
	int cv_type;
	template<int idxR, int idxG, int idxB> cv::Mat& convertImage_impl(cv::Mat const &img);

 public:
	BucketerLut(int hist_size, int lut_nbits):
		BucketerAbstract(hist_size), lut_tobin(8-lut_nbits), lut_size(1<<lut_nbits),
		lut_size2(lut_size*lut_size)
	{
		switch (sizeof(T)) {
			case 1: cv_type = CV_8S; if (hist_size > 127) std::cerr << "BucketerLut::BucketerLut: hist_size too big" << std::endl; break;
			case 2: cv_type = CV_16S; if (hist_size > 32767) std::cerr << "BucketerLut::BucketerLut: hist_size too big" << std::endl; break;
			case 3: cv_type = CV_32S; break;
		}
		int size = color ? lut_size2*lut_size : lut_size;
		lut = new T[size];
	}
	~BucketerLut() { delete[] lut; }

	virtual cv::Mat& convertImage(cv::Mat const &img);
};




/**
 Simple bucketer for gray images: sample by shifting
*/
class BucketerGray: public BucketerAbstract
{
 protected:
	int hist_tobin;  ///< number of bits to shift to get the histogram bucket index
 public:
	virtual BucketerType type() { return btGray; }
	BucketerGray(int hist_nbits):
		BucketerAbstract(1<<hist_nbits), hist_tobin(8-hist_nbits)
		{}
	virtual cv::Mat& convertImage(cv::Mat const &img);
};



/**
 Bucketer based on hue only for color images, using a LUT
 Using only hue seems too sensitive to white or more generally
 unsaturated colors, BucketerHueLowsat is more robust.
*/
template<bool mask = false>
class BucketerHue: public BucketerLut<unsigned char, true, mask>
{
 protected:
 public:
	virtual BucketerType type() { return btHue; }
	BucketerHue(int hist_size, int lut_nbits);
};



/**
 Bruteforce hue bucketer for color images, to compare how much precision
 we lose with the LUT
*/
class BucketerHueBruteforce: public BucketerAbstract
{
 protected:
	int hist_tobin;  ///< number of bits to shift to get the histogram bucket index
	template<int idxR, int idxG, int idxB> cv::Mat& convertImage_impl(cv::Mat const &img);
 public:
	virtual BucketerType type() { return btHueBruteforce; }
	BucketerHueBruteforce(int hist_nbits):
		BucketerAbstract(1<<hist_nbits), hist_tobin(8-hist_nbits)
		{}
	virtual cv::Mat& convertImage(cv::Mat const &img);
};


/**
 Bucketer based on hue only for color images, using a LUT,
 with a special case for very low saturation or low value
 to avoid hue instability.
*/
template<bool includelowsat, bool mask>
class BucketerHueLowsat: public BucketerLut<unsigned char, true, mask>
{
 protected:
	double minsat;
	int minval;
 public:
	virtual BucketerType type() { return includelowsat ? btHueLowsat : btHueHighsat; }
	BucketerHueLowsat(int hist_size, int lut_nbits, int minsat = 32, int minval = 16);
	virtual double param(int i) { switch (i) { case 0: return minsat*255.; case 1: return minval; default: return 0.0; } }
};


/**
 Bucketer based on hue and value for color images, using a LUT.
 Hue and value are not combined, but exclusive depending on saturation and value,
 because low saturation or low value lead to unstable hue.
 It doesn't seem to work better than HueLowsat
*/
template <bool mask = false>
class BucketerHueValue: public BucketerLut<unsigned char, true, mask>
{
 protected:
 public:
	virtual BucketerType type() { return btHueValue; }
	BucketerHueValue(int hist_hue_size, int hist_value_size, int lut_nbits);
};


/**
 Bucketer directly based on RGB values.
 For comparison purpose.
*/
class BucketerRGB: public BucketerAbstract
{
 protected:
	int hist_tobin;
	int hist_size2;
	template<int idxR, int idxG, int idxB> cv::Mat& convertImage_impl(cv::Mat const &img);
	
 public:
	virtual BucketerType type() { return btRGB; }
	BucketerRGB(int hist_nbits):
		BucketerAbstract(1<<hist_nbits), hist_tobin(8-hist_nbits), hist_size2(hist_size*hist_size)
	{
		int hist_Size = 1<<(3*hist_nbits);
		if (hist_Size > 65536) std::cerr << "BucketerRGB::BucketerRGB: hist_size too big" << std::endl;
	}

	virtual cv::Mat& convertImage(cv::Mat const &img);
};

} // namespace histodet

#endif
