#ifndef _DETECTOR_HISTOGRAM_HPP_
#define _DETECTOR_HISTOGRAM_HPP_

#include <fstream>
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#if BUILD_DEMO
#include <opencv2/highgui/highgui.hpp>
#endif

#include "histodet/stretcher.hpp"
#include "histodet/bucketer.hpp"

namespace histodet {

// detector fine setup
const float min_sigma_coeff = 1e-2; // learn

const float min_val_coeff = 0.1; // learn = detect
const float min_support_hist = 2.; // FIXME should it be learnt?



#ifdef CHRONO
#include <sys/time.h>
#include <iostream>
#endif

#if defined SSE && defined __SSE__
const union { int i[4]; __m128 v; } sign_mask = {0x7fffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff};
#endif

#define DetectorHistogram_TEMPLATE(return_type) template <class HistogramMaker, class Distance> return_type DetectorHistogram<HistogramMaker, Distance>

/**
 * Detector based on histogram comparison.
 */
template <class HistogramMaker, class Distance>
class DetectorHistogram
{
 private:
	HistogramMaker hm;
	Stretcher *hs;
	BucketerAbstract *bucketer;

	// learn
	int count_images;
	cv::Mat hist;
	cv::Mat histvar;
	cv::Mat tmp_hist;
	int hist_nbits;
	int hist_size;
	double min_sigma;
	int a, b, min_support;

	// detect
	cv::Mat tpl_hist, tpl_histsig;
	cv::Mat img_res;
	int detect_size;
	float threshold;
	bool dist_inverted;
	bool do_stretch;
	bool do_stretch_normalize;

	bool remove_background;
	cv::Mat mask, warped_mask;
	double mask_ori;
	double mask_size;

 public:
	/**
	 * Constructor
	 * @param hist_nbits number of bits of histogram size
	 * @param bucketer the bucketer used for building histograms
	 */
	DetectorHistogram(int hist_nbits, BucketerAbstract *bucketer);
	/**
	 * Destructor
	 */
	~DetectorHistogram();
	/**
	 * Configure histogram stretching.
	 * @param do_stretch stretch this histogram
	 * @param do_stretch_normalize normalize after stretching, because interpolation artificially increases the number of pixels. But sometimes it works better without...
	 * @param min_val_coeff minimum value a bin must have to be considered non null, as a ratio of the average value of bins
	 * @param min_support_hist minimum support (number of non null bins) of the histogram to accept to stretch and match it, as a ratio of the histogram size
	 */
	void configureStretching(bool do_stretch, bool do_stretch_normalize = true, float min_val_coeff = 0.1, float min_support_hist = 0.5);

	/**
	 * Initialize learning procedure
	 */
	void learnBegin(std::string maskname);
	/**
	 * Add an image to learn, containing an object
	 * @param image the image
	 * @param rect the location of the object to detect
	 */
	void learnImage(cv::Mat const& image, cv::Rect const& rect, double size = 0.0, double ori = 0.0, double center_x = 0.0, double center_y = 0.0);
	/**
	 * Finish learning procedure
	 * @param nameprefix filename prefix for saving the learning results
	 */
	void learnEnd(std::string nameprefix);

	/**
	 * Initialize detection procedure
	 * @param histname the learned histogram description file
	 * @return true if succeeded
	 */
	bool detectInit(std::string histname);
	/**
	 * Detect the object in an image
	 * @param img the image
	 * @param detect_size the size of the object to detect (window size)
	 * @return an image of double representing the score (distance) for each pixel
	 */
	cv::Mat& detectImage(cv::Mat const& img, int detect_size);

	/**
	 * @param[out] dist_inverted whether the distance is inverted or not
	 * @return the threshold read in the histogram file
	 */
	float getThreshold(bool & dist_inverted) { dist_inverted = this->dist_inverted; return threshold; }

	public:
		static void warpPatch(cv::Mat const& patch, cv::Mat &warpedPatch, double angle, double scale)
		{
			cv::Point2f center(patch.cols/2., patch.rows/2.);
			cv::Size size = patch.size();
			cv::Mat M = cv::getRotationMatrix2D(center, angle, scale);
			cv::warpAffine(patch, warpedPatch, M, size, cv::INTER_NEAREST, cv::BORDER_CONSTANT, 0);
		}

};




///*********************************************************************************************************************************************
/// Template functions
///*********************************************************************************************************************************************


DetectorHistogram_TEMPLATE()::DetectorHistogram(int hist_nbits, BucketerAbstract *bucketer):
	hm(1<<hist_nbits, bucketer), hs(NULL),
	bucketer(bucketer),
	hist_nbits(hist_nbits),
	hist_size(1<<hist_nbits),
	do_stretch(false), do_stretch_normalize(false),
	remove_background(false)
{}


DetectorHistogram_TEMPLATE()::~DetectorHistogram()
{}

DetectorHistogram_TEMPLATE(void)::configureStretching(bool do_stretch, bool do_stretch_normalize, float min_val_coeff, float min_support_hist)
{
	// FIXME should min_support_hist be learnt? it may decrease generalization...
	if (do_stretch)
	{
		this->do_stretch = true;
		this->do_stretch_normalize = do_stretch_normalize;
		if (hs) delete hs;
		hs = new Stretcher(hist_nbits, std::floor((hist_size/min_support_hist)+0.5), do_stretch_normalize);
	}
}


DetectorHistogram_TEMPLATE(void)::learnBegin(std::string maskname)
{
	hist = cv::Mat::zeros(hist_size, 1, CV_64F);
	histvar = cv::Mat::zeros(hist_size, 1, CV_64F);
	tmp_hist = cv::Mat::zeros(hist_size, 1, CV_64F);
	min_sigma = min_sigma_coeff * 1./hist_size;
	count_images = 0;
	min_support = hist_size;

#if BUILD_DEMO
	if (maskname.size() >= 2)
	{
		remove_background = true;
		
		size_t last = maskname.rfind("/");
		if (last == std::string::npos) last = 0;
		std::string path = maskname.substr(0, last);

		cv::FileStorage fs(maskname, cv::FileStorage::READ);
		if (!fs.isOpened()) { std::cerr << "DetectorHistogram::learnBegin: failed to load mask file " << maskname << std::endl; return; }
		std::string name;
		fs["mask_file"] >> name; name = path + "/" + name;
		if (!(mask = cv::imread(name.c_str(), 1)).data)
			{ std::cout << "could not load file " << name << std::endl; return; }
		fs["orientation"] >> mask_ori;
		fs["size"] >> mask_size;

		warped_mask = cv::Mat(mask.rows, mask.cols, CV_8UC1);
	}
#endif
}


DetectorHistogram_TEMPLATE(void)::learnImage(cv::Mat const& img, cv::Rect const& rect, double size, double ori, double center_x, double center_y)
{
	// TODO ensure rect is square
	if (remove_background)
	{
		warpPatch(mask, warped_mask, ori-mask_ori, size/mask_size);
		cv::Mat sub_img = img(rect);
		//cv::Mat sub_mask = warped_mask(cv::Rect((warped_mask.rows-rect.height)/2, (warped_mask.cols-rect.width)/2, rect.width, rect.height));
		cv::Mat sub_mask = warped_mask(cv::Rect(warped_mask.cols/2-(center_x-rect.x), warped_mask.rows/2-(center_y-rect.y), rect.width, rect.height));
		sub_img &= sub_mask;
	}
	
	hm.loadImage(img, 1);
	float min_val = (rect.width*rect.height) / (float)hist_size * min_val_coeff;
	bool zap;
	cv::Mat &h = hm.get(rect);
	if (do_stretch)
	{
		hs->computeSupport(h, min_val, a, b);
		if (b-a+1 < min_support) min_support = b-a+1;
	}
	cv::Mat &hstr = do_stretch ? hs->stretch(h, min_val, zap) : h;
	hstr.convertTo(tmp_hist, CV_64F, 1./(rect.width*rect.height), 0.);
	hist += tmp_hist;
	histvar += tmp_hist.mul(tmp_hist);
	++count_images;
}


DetectorHistogram_TEMPLATE(void)::learnEnd(std::string nameprefix)
{
	hist /= count_images;
	histvar = histvar / count_images - hist.mul(hist);

	cv::Mat histf;
	cv::FileStorage fs(nameprefix+".xml", cv::FileStorage::WRITE);
	hist.convertTo(histf, CV_32F);
	fs << "hist" << histf;
	for(int i = 0; i < histvar.rows; ++i)
	{
		histvar.at<double>(i,0) = sqrt(histvar.at<double>(i,0));
		if (histvar.at<double>(i,0) < min_sigma) histvar.at<double>(i,0) = min_sigma; // prevent div by 0 on detect and overfitting (TODO laplacian smoothing ?)
	}
	histvar.convertTo(histf, CV_32F);
	fs << "hist_sig" << histf;
	fs << "do_stretch" << do_stretch;
	fs << "do_stretch_normalize" << do_stretch_normalize;
	fs << "min_support" << min_support;
	fs << "bucketer_type" << bucketer->type();
	fs << "bucketer_param1" << bucketer->param(0);
	fs << "bucketer_param2" << bucketer->param(1);
	fs << "threshold" << (double)0.0;
	fs.release();

	// save also hist to easily plot it
	std::ofstream f((nameprefix+".dat").c_str());
	for(int i = 0; i < hist_size; ++i)
		f << i << "\t" << hist.at<double>(i,0) << "\t" << histvar.at<double>(i,0) << std::endl;
	f.close();
}

DetectorHistogram_TEMPLATE(bool)::detectInit(std::string histname)
{
	cv::FileStorage fs(histname, cv::FileStorage::READ);
	if (!fs.isOpened()) { std::cerr << "DetectorHistogram::detectInit: failed to load histogram file " << histname << std::endl; return false; }
	fs["hist"] >> tpl_hist;
	if (hist_size != tpl_hist.rows) { std::cerr << "DetectorHistogram::detectInit: hist_size not consistent" << std::endl; return false; }
	fs["do_stretch"] >> do_stretch;
	fs["do_stretch_normalize"] >> do_stretch_normalize;
	//fs["min_support"] >> min_support;
	fs["hist_sig"] >> tpl_histsig;
	fs["threshold"] >> threshold;
	fs.release();

	configureStretching(do_stretch, do_stretch_normalize);

	threshold = Distance::transform_threshold(threshold, detect_size*detect_size, hist_size, dist_inverted);

	hm = HistogramMaker(hist_size, bucketer);
	return true;
}


DetectorHistogram_TEMPLATE(cv::Mat&)::detectImage(cv::Mat const& img, int detect_size)
{
	if (img.cols != img_res.cols || img.rows != img_res.rows)
	{
		bool invert;
		img_res = cv::Mat::ones(img.rows, img.cols, CV_32F)
		          * Distance::transform_threshold(2.0, detect_size*detect_size, hist_size, invert);
	}

	hm.loadImage(img, detect_size);

	float dsds = detect_size*detect_size;
	cv::Mat scaled_tpl_hist = tpl_hist * dsds;
	cv::Mat scaled_tpl_histsig = tpl_histsig * dsds;
	float min_val = dsds / hist_size * min_val_coeff;
	bool zap = false;

#ifdef CHRONO
	static struct timeval tv1, tv2;
	static double avgt = 0., mint = 1e6;
	static int n = 0;
	gettimeofday(&tv1,NULL);
#endif
	
	int x, y;
	while (true)
	{
		cv::Mat &histref = hm.getNext(x,y);
		if (x < 0) break;
		
		cv::Mat &hist_stretch = do_stretch ? hs->stretch(histref, min_val, zap) : histref;
		*((float*)(img_res.data+img_res.step[0]*y)+x) = (zap ? Distance::infinite() : Distance::dist(hist_stretch, scaled_tpl_hist, scaled_tpl_histsig));
//std::cout << "histo score " << *((float*)(img_res.data+img_res.step[0]*y)+x) << std::endl;
	}

#ifdef CHRONO
	gettimeofday(&tv2,NULL);
	double t = (tv2.tv_sec-tv1.tv_sec) + (tv2.tv_usec-tv1.tv_usec)*1e-6;
	avgt += t; if (t < mint) mint = t; ++n;
	std::cout << "[HistogramDetection] Processed image in " << t << " seconds (average " << avgt/n << " , min " << mint << ")" << std::endl;
#endif
	
	return img_res;
}


} // namespace histodet


#endif

