#ifndef _HISTOGRAM_ABSTRACT_HPP_
#define _HISTOGRAM_ABSTRACT_HPP_

#include <opencv2/imgproc/imgproc.hpp>

#include "histodet/bucketer.hpp"

/*
 * Methods are not virtual because inherited classes should be used as template and methods inlined
 */

namespace histodet {

/**
 * This class represents a histogram, which is basically a cv::Mat, except
 * that it allows for a bin at index -1, which is ignored but which should
 * not provoke memory corruption.
 */
class Hist
{
 private:
	cv::Mat hist_data;
	cv::Mat hist;
 public:
	Hist() {}
	Hist(size_t size) { build(size); }
	void build(size_t size)
	{
		// we shift 4 floats to preserve alignment for sse
		hist_data = cv::Mat::zeros(size+4, 1, CV_32F);
		hist = cv::Mat(hist_data, cv::Rect(0, 4, 1, size));
	}
	inline cv::Mat & operator()() { return hist; }
	inline operator cv::Mat&() { return hist; }
	
	inline void clear() { hist_data.setTo((float)0.); }
	inline float* data() { return (float*)hist.data; }
	inline int getExcluded() { return data()[-1]; }

};

/**
 * Abstract class for computing an histogram, sequentially or randomly.
 */
class HistogramAbstract
{
 protected:
 	BucketerAbstract *bucketer;
	cv::Mat img;     ///< temporary storage of the image to process
	int hist_size;   ///< number of buckets of the histogram
	int win_size;    ///< histogram window size
	int win_hsize;   ///< histogram window half size
	int cur_x, cur_y; ///< current position when scanning
	Hist hist;       ///< the histogram
	bool first_line; ///< are we processing the first line
	bool first_col;  ///< are we processing the first column
	bool finished;   ///< have we finished processing the whole image
protected:
	inline void updatePos(int &pos_x, int &pos_y);
	void setWinSize(int window_size)
		{ win_hsize = window_size/2; win_size = win_hsize*2+1; }
 public:
	/**
	 * Constructor
	 * @param hist_size the histogram size
	 * @param bucketer the bucketer to convert a pixel into a bucket index
	 */
	HistogramAbstract(int hist_size, BucketerAbstract *bucketer);
	/**
	 * Destructor
	 */
	~HistogramAbstract();
	/**
	 * Load the image from which we will then compute histograms overs some areas
	 * @param image the image
	 * @param window_size the window size over which histograms will be computed when scanning
	 */
	void loadImage(cv::Mat const &image, int window_size);
	/**
	 * Compute the histogram of an arbitrary area in the image
	 */
	inline cv::Mat& get(cv::Rect const& rect) { return hist; }
	/**
	 * Sequential computation of histogram: compute at next position
	 * @param[out] pos_x x position of the window center of the returned result
	 * @param[out] pos_y y position of the window center of the returned result
	 * @return the histogram over the window centered at pos_x,pos_y
	 */
	inline cv::Mat& getNext(int &pos_x, int &pos_y) { return hist; }
	/**
	 * Return the number of pixels excluded from the histogram
	 */
	inline int getExcluded() { return hist.getExcluded(); }
};



///*********************************************************************************************************************************************
/// Inlined functions
///*********************************************************************************************************************************************


void HistogramAbstract::updatePos(int &pos_x, int &pos_y)
{
	// res
	pos_x = cur_x;
	pos_y = cur_y;

	// update cur pos
	++cur_x;
	//first_col = false;
	if (cur_x >= img.cols - win_hsize)
	{
		cur_x = win_hsize;
		first_line = false;
		first_col = true;
		++cur_y;
		if (cur_y >= img.rows - win_hsize)
		{
			cur_x = cur_y = -1;
			finished = true;
		}
	}

}


} // namespace histodet


#endif

