#ifndef _REFINER_CORRELATION_HPP_
#define _REFINER_CORRELATION_HPP_

#if BUILD_DEMO
#include <opencv2/highgui/highgui.hpp>
#endif

#include "histodet/refinerAbstract.hpp"
#include "histodet/hsv.hpp"

#define PATCHLOWSAT 0

#ifdef CHRONO
#include <sys/time.h>
#include <iostream>
int ncorrel;
#endif

namespace histodet {

	/**
	 *
	 * TODO:
	 * - try SSD instead of SAD because large differences mean totally different hues..
	 * - precompute warped templates used for initial brute force search
	 * - dynamically compute final steps so that it represents less than 1 pix wrt the template size
	 * - deal correctly with warped_templat size (rotation and resize)
	 * - don't recompute rotation when only changing scale ?
	 * - reduce roi when refining
	 * - maybe instead of completely masking low sat, apply a coefficient in correl computation?
	 */
	class RefinerCorrelation: public RefinerAbstract
	{
		public:
			struct Templat { cv::Mat templat, mask; };
		private:
			bool rgb_order;
			Templat init_templat;
			Templat warped_templat;
			double threshold; ///< correlation threshold to have positive detection
			double size; ///< size in pixels of the object in the template
			double orientation; ///< orientation of the object in the template

		private:
			
			/**
			 * Convert a RGB or BGR image into hue image
			 * (template parameters describe whether it is RGB or BGR)
			 * @param color_img input rgb or bgr color image
			 * @param hue_img output hue image (must be initialized to CV_8UC1 and same size than color_img)
			 */
			template<int idxR, int idxG, int idxB>
			void convertHue(cv::Mat const& color_img, cv::Mat & hue_img)
			{
				int channels = color_img.channels();
				for(int y = 0; y < color_img.rows; ++y)
				for(int x = 0; x < color_img.cols; ++x)
				{
					unsigned char *pix = color_img.data+color_img.step[0]*y+x*channels;
					hue_img.ptr<unsigned char>(y)[x] =
						rgb2h(*(pix+idxR), *(pix+idxG), *(pix+idxB));
				}
			}

			/**
			 * Convert a RGB or BGR image into hue image, but remove from the mask pixels for which saturation is too low
			 */
			template<int idxR, int idxG, int idxB>
			void convertHueLowSat(cv::Mat const& color_img, cv::Mat & hue_img, cv::Mat & mask)
			{
				int channels = color_img.channels();
				for(int y = 0; y < color_img.rows; ++y)
				for(int x = 0; x < color_img.cols; ++x)
				{
					float h,s,v;
					unsigned char *pix = color_img.data+color_img.step[0]*y+x*channels;
					rgb2hsv<256>(*(pix+idxR), *(pix+idxG), *(pix+idxB), h, s, v);
					hue_img.ptr<unsigned char>(y)[x] = h;
// TODO make homogeneous
					if (s < 0.25 || v < 64) mask.ptr<unsigned char>(y)[x] = 0;
				}
			}

			
			/**
			 * Warp template
			 */
			void warpTemplate(Templat const& templat, Templat &warpedTemplat, double angle, double scale)
			{
				cv::Point2f center(templat.templat.cols/2., templat.templat.rows/2.);
				cv::Size size = templat.templat.size();
				cv::Mat M = getRotationMatrix2D(center, angle, scale);
				// should not interpolate template, because hue is periodic
				// if we want interpolation we should rotate the RGB image and convert again to hue
				// this is not too expensive as we do it only once during initialization, but let's try simple first
				cv::warpAffine(templat.templat, warpedTemplat.templat, M, size, cv::INTER_NEAREST, cv::BORDER_CONSTANT, 0);
				cv::warpAffine(templat.mask, warpedTemplat.mask, M, size, cv::INTER_LINEAR, cv::BORDER_CONSTANT, 0);
			}
			
			/**
			 * Absolute value
			 */
			template<typename T> inline T abs(T x) { return x>0?x:-x; }

			/**
			 * Periodic absolute difference
			 */
			template<typename R, typename T, int maxval>
			inline R padiff(T a, T b)
			{
				int d = abs<int>(a-b);
				if (d > maxval/2) d = maxval-d;
				return d;
			}
			
			/**
			 * SAD correlation variant for periodic values (like hue) with mask management
			 * @param templat
			 * @param mask
			 * @param patch
			 * @param shift
			 * @return correlation score between 0 (identical) and 1 (different)
			 */
#if PATCHLOWSAT
			double correl(Templat const& templat, Templat const& patch, cv::Point const& shift)
#else
			double correl(Templat const& templat, cv::Mat const& patch, cv::Point const& shift)
#endif
			{
				int count = 0;
				int score = 0;

				for(int y = 0; y < templat.templat.rows; ++y)
				{
					const unsigned char* templat_line = templat.templat.ptr<unsigned char>(y);
					const unsigned char* mask_line = templat.mask.ptr<unsigned char>(y);
#if PATCHLOWSAT
					const unsigned char* patch_line = patch.templat.ptr<unsigned char>(y+shift.y);
					const unsigned char* patchmask_line = patch.mask.ptr<unsigned char>(y+shift.y);
#else
					const unsigned char* patch_line = patch.ptr<unsigned char>(y+shift.y);
#endif
					
					for(int x = 0; x < templat.templat.cols; ++x)
					{
#if PATCHLOWSAT
						if (mask_line[x] == 0 && patchmask_line[x+shift.x] != 0) { ++count; score += 127; } else
						if (mask_line[x] != 0 && patchmask_line[x+shift.x] == 0) { ++count; score += 127; } else
						if (mask_line[x] != 0 && patchmask_line[x+shift.x] != 0) { ++count; score += padiff<int, unsigned char, 256>(templat_line[x], patch_line[x+shift.x]); } else
							++count;
#else
						if (mask_line[x] == 0) continue;
						++count;
						score += padiff<int, unsigned char, 256>(templat_line[x], patch_line[x+shift.x]);
#endif
					}
				}
				return (double)score/((double)count*127.0);
			}

#if PATCHLOWSAT
			double exploreTranslation(Templat const& templat, Templat const &patch, cv::Rect roi, int step, cv::Point &res)
#else
			double exploreTranslation(Templat const& templat, cv::Mat const &patch, cv::Rect roi, int step, cv::Point &res)
#endif
			{
				double best_score = 1.0;
				for(int x = roi.x; x <= roi.x+roi.width -templat.templat.cols ; x += step)
				for(int y = roi.y; y <= roi.y+roi.height-templat.templat.rows; y += step)
				{
#ifdef CHRONO
++ncorrel;
#endif
					cv::Point shift(x,y);
					double score = correl(templat, patch, shift);
					if (score < best_score) { best_score = score; res = shift; }
				}
				return best_score;
			}

			
			
		public:

			/**
			 * Constructor
			 * @param templat the template for correlation
			 * @param mask CV_8UC1 and same size than templat, non zero pixels indicate corresponding object pixels in templat
			 * @param rgb_order true if templat has values in order RGB, false if BGR
			 */
			RefinerCorrelation(std::string correlname, bool rgb_order = false):
				rgb_order(rgb_order)
			{
#if BUILD_DEMO // FIXME we need to do this even if it is not a demo if we want to use this class
				// load file
				size_t last = correlname.rfind("/");
				if (last == std::string::npos) last = 0;
				std::string path = correlname.substr(0, last);
				
				cv::FileStorage fs(correlname, cv::FileStorage::READ);
				if (!fs.isOpened()) { std::cerr << "RefinerCorrelation::RefinerCorrelation: failed to load correlation file " << correlname << std::endl; return; }
				std::string name;
				cv::Mat templat, mask;
				fs["template"] >> name; name = path + "/" + name;
				if (!(templat = cv::imread(name.c_str(), -1)).data)
					{ std::cout << "could not load file " << name << std::endl; return; }
				fs["template_mask"] >> name; name = path + "/" + name;
				if (!(mask = cv::imread(name.c_str(), 0)).data)
					{ std::cout << "could not load file " << name << std::endl; return; }
				fs["threshold"] >> threshold;
				fs["size"] >> size;
				fs["orientation"] >> orientation;
	
				// init
				init_templat.templat = cv::Mat(templat.rows, templat.cols, CV_8UC1);
				init_templat.mask = mask.clone();
				warped_templat.templat = cv::Mat(templat.rows, templat.cols, CV_8UC1);
				warped_templat.mask = cv::Mat(templat.rows, templat.cols, CV_8UC1);
				
				// convert rgb to hue
				if (rgb_order) convertHueLowSat<0,1,2>(templat, init_templat.templat, init_templat.mask);
									else convertHueLowSat<2,1,0>(templat, init_templat.templat, init_templat.mask);
// cv::imwrite("init_template.templat.png", init_templat.templat);
// cv::imwrite("init_template.mask.png", init_templat.mask);
#endif
			}


			/**
			 * Refine with correlation
			 *
			 */
			bool refine(Object &object, cv::Mat const& image, double &correl_score)
			{
#ifdef CHRONO
	static struct timeval tv1, tv2;
	static double avgt = 0., mint = 1e6;
	static int n = 0;
	ncorrel = 0;
	gettimeofday(&tv1,NULL);
#endif

				double angleStep = 90.0/8; // 90.0/8
				double scaleStep = 1.1; // 1.1;
				const double maxScale = 1.4;
				const double angleAccuracy = 1.0;
				const double scaleAccuracy = 1.01;
				const int exhaustive_res = 2; // pixels
				const double refine_size_ratio = 3.0;
				const int refine_size_min = 10; // pixels

				cv::Point offset;
				cv::Mat patch = extractPatch(object, image, offset);
				
				// convert rgb to hue
#if PATCHLOWSAT
				Templat hue_patch;
				hue_patch.templat = cv::Mat(patch.rows, patch.cols, CV_8UC1);
				hue_patch.mask = cv::Mat(patch.rows, patch.cols, CV_8UC1, 255);
				if (rgb_order) convertHueLowSat<0,1,2>(patch, hue_patch.templat, hue_patch.mask);
									else convertHueLowSat<2,1,0>(patch, hue_patch.templat, hue_patch.mask);
#else
				cv::Mat hue_patch = cv::Mat(patch.rows, patch.cols, CV_8UC1);
				if (rgb_order) convertHue<0,1,2>(patch, hue_patch);
									else convertHue<2,1,0>(patch, hue_patch);
#endif
				
// cv::imwrite("whole_image.png", image);
// cv::imwrite("patch.png", patch);
// cv::imwrite("hue_patch.png", hue_patch);


// std::cout << "Start Large Grained Brute Force Search" << std::endl;
				// large grained brute force search on orientation and size
				cv::Rect roi(0, 0, patch.cols, patch.rows);
				cv::Point pos, best_pos;
				double best_a, best_s, best_score = 2.0;
				for(double a = 0; a < 359.5; a += angleStep)
				{
					double step = scaleStep, jump = 1.0;
					for(double s = 1.0; s < maxScale; jump = step/jump, s *= jump, step = 1.0/step)
					{
						warpTemplate(init_templat, warped_templat, a, s);

// std::ostringstream oss; oss << "warped_" << a << "_" << s;
// cv::imwrite(oss.str()+"_hue.png", warped_templat.templat);
// cv::imwrite(oss.str()+"_mask.png", warped_templat.mask);
						
						double score = exploreTranslation(warped_templat, hue_patch, roi, exhaustive_res, pos);
						if (score < best_score) { best_a = a; best_s = s; best_pos = pos; best_score = score; }
// std::cout << "correl for a " << a << " s " << s << " score " << score << " (at " << pos.x << "," << pos.y << ")" << std::endl;
					}
				}

#ifdef CHRONO
				int ncorrel1 = ncorrel;
				ncorrel = 0;
#endif
				
// std::cout << "# Best is a " << best_a << " s " << best_s << " score " << best_score << " (at " << best_pos.x << "," << best_pos.y << ") -- Start refining" << std::endl;

				roi.width = best_s*size/refine_size_ratio; if (roi.width < refine_size_min) roi.width = refine_size_min; roi.height = roi.width;
				roi.x = best_pos.x - roi.width/2;
				roi.y = best_pos.y - roi.height/2;
				roi.width += warped_templat.templat.cols; roi.height += warped_templat.templat.rows;
				if (roi.x < 0) roi.x = 0;
				if (roi.y < 0) roi.y = 0;
				if (roi.x + roi.width  > patch.cols) roi.width  = patch.cols - roi.x;
				if (roi.y + roi.height > patch.rows) roi.height = patch.rows - roi.y;

// std::cout << "Refinement roi is " << roi.x << "," << roi.y << "," << roi.width << "," << roi.height << std::endl;
				
				// keep refining orientation and size
				while (angleStep > angleAccuracy || scaleStep > scaleAccuracy)
				{
					angleStep /= 2.0;
					scaleStep = (1.0+scaleStep)/2.0;
					double b_a = best_a, b_s = best_s;
// std::cout << "search around a " << b_a << " s " << b_s << " best_score " << best_score << std::endl;
					
					for(int ia = -1; ia <= 1; ++ia)
					for(int is = -1; is <= 1; ++is)
					{
						if (ia == 0 && is == 0) continue;
						double a = b_a + ia*angleStep;
						double s = (is < 0 ? b_s/scaleStep : (is > 0 ? b_s*scaleStep : b_s));
						warpTemplate(init_templat, warped_templat, a, s);
						double score = exploreTranslation(warped_templat, hue_patch, roi, 1, pos);
						if (score < best_score) { best_a = a; best_s = s; best_pos = pos; best_score = score; }
// std::cout << "correl for a " << a << " s " << s << " score " << score << " (at " << pos.x << "," << pos.y << ")" << std::endl;
					}
				}
// std::cout << "## Best is a " << best_a << " s " << best_s << " score " << best_score << " (at " << best_pos.x << "," << best_pos.y << ")" << std::endl;

#ifdef CHRONO
	gettimeofday(&tv2,NULL);
	double t = (tv2.tv_sec-tv1.tv_sec) + (tv2.tv_usec-tv1.tv_usec)*1e-6;
	avgt += t; if (t < mint) mint = t; ++n;
	std::cout << "[ObjectRefiner] Processed image in " << t << " seconds (average " << avgt/n << " , min " << mint << ") ("
		<< ncorrel1+ncorrel << " correls, " << ncorrel1 << " for initial exhaustive search, " << ncorrel << " for refinement)" << std::endl;
#endif

				correl_score = best_score;
				object.score = 1.-best_score;
		
				if (best_score > threshold)
				{
// std::cout << "## OBJECT REJECTED with score " << best_score << std::endl;
					return false;
				}
				else
				{
					object.set(Object::lpPosU, offset.x + best_pos.x + warped_templat.templat.cols/2, 2.0);
					object.set(Object::lpPosV, offset.y + best_pos.y + warped_templat.templat.rows/2, 2.0);
					object.set(Object::lpSize, best_s*size, scaleAccuracy*best_s*size);
					double h = (orientation+best_a)*M_PI/180.; if (h > M_PI) h -= 2*M_PI;
					object.set(Object::lpHeading, h, angleAccuracy*M_PI/180.);
					
// std::cout << "## OBJECT FOUND at " << object.loc[Object::lpPosU] << "," << object.loc[Object::lpPosV] << " size " << object.loc[Object::lpSize] << " heading " << object.loc[Object::lpHeading]*180./M_PI << " score " << best_score << std::endl;
					return true;
				}

			}


	};






} // namespace histodet


#endif

