#ifndef _REFINER_HUEBLOBS_HPP_
#define _REFINER_HUEBLOBS_HPP_

#include <vector>
#include <set>

#include "histodet/refinerAbstract.hpp"
#include "histodet/distance.hpp"

//#define CHRONO
#define DISPLAY 0

#if DISPLAY
#include <iomanip>
#include <opencv2/highgui/highgui.hpp>
#endif

#ifdef CHRONO
#include <sys/time.h>
#include <iostream>
#endif


namespace histodet {

	struct Blob {
		double hue;
		double x, y, asize, size; // constant parameters
		double adist, dist, aangle, angle; // variable parameters (absolute and relative values)
		bool operator<(Blob const & blob) const { return asize > blob.asize; }
	};

	struct Solution {
		std::vector<Blob> blobs;
		double bary_x, bary_y;
		double size;
		double score;
	};

	
	class RefinerHueBlobs: public RefinerAbstract
	{
		public:
			typedef DistanceL2n Distance;
			
		private:
			bool rgb_order;
			double threshold; ///< correlation threshold to have positive detection
			cv::Mat hue_patch;
			unsigned count_images;

			uint nmclusters; ///< number of model clusters
			cv::Mat model; ///< [cluster1_hue, cluster1_dist, cluster1_angle, cluster1_size, ...] ; used for comparison
			cv::Mat model_sig; ///< standard deviations of model
			cv::Mat ratios; ///< [cluster1_size, cluster1_angle, cluster1_centerX, cluster1_centerY ...] ; used to compute global size and heading, and center
			cv::Mat ratios_sig; ///< standard deviations of ratios
		protected:
		
		private:
			/**
			 * Convert a RGB or BGR image into hue image, but remove from the mask pixels for which saturation is too low
			 */
			template<int idxR, int idxG, int idxB>
			cv::Mat createSamples(cv::Mat const& patch)
			{
				hue_patch = cv::Mat(patch.rows, patch.cols, CV_8UC1);
				cv::Mat samples = cv::Mat(0, 3, CV_32FC1);
				cv::Mat sample = cv::Mat(1, 3, CV_32FC1);
				int channels = patch.channels();
				for(int y = 0; y < patch.rows; ++y)
				for(int x = 0; x < patch.cols; ++x)
				{
					float h,s,v;
					unsigned char *pix = patch.data+patch.step[0]*y+x*channels;
					rgb2hsv<256>(*(pix+idxR), *(pix+idxG), *(pix+idxB), h, s, v, 60.0);
					hue_patch.at<uchar>(y,x) = h;

					if (s >= 0.2 && v > 16)
					{
						sample.at<float>(0,0) = y+0.5;
						sample.at<float>(0,1) = x+0.5;
						sample.at<float>(0,2) = h;
						samples.push_back(sample);
					}
				}
				return samples;
			}

			template<typename T> inline void constraint_angle(T angle_comp, T & angle_fix) { if (angle_comp > M_PI) angle_fix -= 2*M_PI; else if (angle_comp < -M_PI) angle_fix += 2*M_PI; }
			void floodFillExtract(cv::Mat const & image, std::list<Blob> & blobs);
			void extractClusters(cv::Mat & patch, std::vector<Blob> & blobs);
			void evalBlobs(std::vector<int> & assigned, std::vector<Blob> & blobs, double &bary_x, double &bary_y, double &size);
			void match(std::vector<int> & assigned, std::set<int> & remaining, std::vector<Blob> & blobs, Solution & best_solution);
			void computeState(Object &object, Solution & best_solution, cv::Point offset);

		
		public:
			RefinerHueBlobs(bool rgb_order = false): rgb_order(rgb_order) {}
			void learnBegin(unsigned nclusters);
			void learnImage(cv::Mat const& img, cv::Rect const& rect, double &size, double &ori, double &center_x, double &center_y, std::vector<std::pair<double,double> > &mblobs);
			void learnEnd(std::string nameprefix);
			bool refineInit(std::string descriptorname);
			virtual bool refine(Object &object, cv::Mat const& image, double &distance);

	};


	///*********************************************************************************************************************************************
	/// learnBegin
	///*********************************************************************************************************************************************
	
	void RefinerHueBlobs::learnBegin(unsigned nclusters)
	{
		nmclusters = nclusters;
		model = cv::Mat::zeros(1, nmclusters*4, CV_32FC1);
		model_sig = cv::Mat::zeros(1, nmclusters*4, CV_32FC1);
		ratios = cv::Mat::zeros(1, nmclusters*4, CV_32FC1);
		ratios_sig = cv::Mat::zeros(1, nmclusters*4, CV_32FC1);
		count_images = 0;
	}

	///*********************************************************************************************************************************************
	/// learnImage
	///*********************************************************************************************************************************************
	
	void RefinerHueBlobs::learnImage(cv::Mat const& img, cv::Rect const& rect, double &size, double &ori, double &center_x, double &center_y, std::vector<std::pair<double,double> > &mblobs)
	{
		ori *= M_PI/180.;
		
		cv::Rect rect2;
		rect2.width  = 1.25*rect.width; rect2.height = 1.25*rect.height;
		rect2.x = rect.x-rect.width/8; rect2.y = rect.y-rect.height/8;
		if (rect2.x < 0) rect2.x = 0; if (rect2.y < 0) rect2.y = 0;
		if (rect2.x + rect2.width  > img.cols) rect2.width  = img.cols - rect.x;
		if (rect2.y + rect2.height > img.rows) rect2.height = img.rows - rect.y;
		cv::Point offset(rect2.x, rect2.y);
		cv::Mat patch(img, rect2);

		std::vector<Blob> blobs;
		extractClusters(patch, blobs);
		if (blobs.size() < mblobs.size())
			{ std::cerr << "RefinerHueBlobs::learnImage: not enough blobs detected" << std::endl; return; }

// std::cout << "found " << blobs.size() << " blobs " << std::endl;
		std::vector<int> assigned(mblobs.size());
		for(int i = 0; i < mblobs.size(); ++i)
		{
			int jclosest;
			double closest_dist2 = 1e10;
			for(int j = 0; j < blobs.size(); ++j)
			{
				double dist2 = sqr(mblobs[i].first-offset.x-blobs[j].x) + sqr(mblobs[i].second-offset.y-blobs[j].y);
				if (dist2 < closest_dist2) { closest_dist2 = dist2; jclosest = j; }
			}
			assigned[i] = jclosest;
// std::cout << "closest from " << i << " at " << mblobs[i].first << "," << mblobs[i].second << " : " << jclosest << " at " << blobs[jclosest].x << "," << blobs[jclosest].y << std::endl;
			for(int k = 0; k < i; ++k) if (assigned[k] == jclosest)
				{ std::cerr << "RefinerHueBlobs::learnImage: used twice same cluster as closest" << std::endl; return; }
		}

		double bary_x, bary_y, nsize;
		evalBlobs(assigned, blobs, bary_x, bary_y, nsize);

		// update mode
		// TODO hue and angle modulo
		// TODO center_x and center_y
		for(int i = 0; i < nmclusters; ++i)
		{
			model.at<float>(0,4*i+0) += blobs[assigned[i]].hue;
			model.at<float>(0,4*i+1) += blobs[assigned[i]].dist;
			model.at<float>(0,4*i+2) += blobs[assigned[i]].angle;
			model.at<float>(0,4*i+3) += blobs[assigned[i]].size;

			model_sig.at<float>(0,4*i+0) += sqr(blobs[assigned[i]].hue);
			model_sig.at<float>(0,4*i+1) += sqr(blobs[assigned[i]].dist);
			model_sig.at<float>(0,4*i+2) += sqr(blobs[assigned[i]].angle);
			model_sig.at<float>(0,4*i+3) += sqr(blobs[assigned[i]].size);
// std::cout << "angle label " << ori*180./M_PI << " blob " << i << " " << blobs[assigned[i]].aangle*180./M_PI << " correction " << (ori - blobs[assigned[i]].aangle)*180./M_PI << std::endl;
			ratios.at<float>(0,4*i+0) += size / blobs[assigned[i]].asize;
			ratios.at<float>(0,4*i+1) += ori - blobs[assigned[i]].aangle;
			ratios.at<float>(0,4*i+2) += 0.0;
			ratios.at<float>(0,4*i+3) += 0.0;
			
			ratios_sig.at<float>(0,4*i+0) += sqr(size / blobs[assigned[i]].asize);
			ratios_sig.at<float>(0,4*i+1) += sqr(ori - blobs[assigned[i]].aangle);
			ratios_sig.at<float>(0,4*i+2) += sqr(0.0);
			ratios_sig.at<float>(0,4*i+3) += sqr(0.0);
		}
		
		++count_images;
	}


	///*********************************************************************************************************************************************
	/// learnEnd
	///*********************************************************************************************************************************************
	
	void RefinerHueBlobs::learnEnd(std::string nameprefix)
	{
		const float minsig = 1e-3;
		model /= count_images;
		model_sig = model_sig / count_images - model.mul(model);
		ratios /= count_images;
		ratios_sig = ratios_sig / count_images - ratios.mul(ratios);
		for(int i = 0; i < model.cols; ++i)
		{
			model_sig.at<float>(0,i) = sqrt(model_sig.at<float>(0,i));
			if (model_sig.at<float>(0,i) < minsig) model_sig.at<float>(0,i) = minsig; // prevent div by 0 and overfitting
			ratios_sig.at<float>(0,i) = sqrt(ratios_sig.at<float>(0,i));
			if (ratios_sig.at<float>(0,i) < minsig) ratios_sig.at<float>(0,i) = minsig; // prevent div by 0 and overfitting
		}
		
		cv::FileStorage fs(nameprefix+".xml", cv::FileStorage::WRITE);
		fs << "model" << model;
		fs << "model_sig" << model_sig;
		fs << "ratios" << ratios;
		fs << "ratios_sig" << ratios_sig;
		fs << "threshold_ref" << (double)0.0;
		fs.release();
	}


	///*********************************************************************************************************************************************
	/// refineInit
	///*********************************************************************************************************************************************
	
	bool RefinerHueBlobs::refineInit(std::string descriptorname)
	{
		cv::FileStorage fs(descriptorname, cv::FileStorage::READ);
		if (!fs.isOpened()) { std::cerr << "RefinerHueBlobs::RefinerHueBlobs: failed to load descriptor file " << descriptorname << std::endl; return false; }
		fs["model"] >> model;
		fs["model_sig"] >> model_sig;
		fs["ratios"] >> ratios;
		fs["ratios_sig"] >> ratios_sig;
		fs["threshold_ref"] >> threshold;
		nmclusters = model.cols/4;
		return true;
	}

	///*********************************************************************************************************************************************
	/// evalBlobs
	///*********************************************************************************************************************************************
	
	void RefinerHueBlobs::evalBlobs(std::vector<int> & assigned, std::vector<Blob> & blobs, double &bary_x, double &bary_y, double &size)
	{
		// barycenter
		bary_x = 0.0; bary_y = 0.0;
		for(int i = 0; i < nmclusters; ++i)
			{ bary_x += blobs[assigned[i]].x; bary_y += blobs[assigned[i]].y; }
		bary_x /= nmclusters; bary_y /= nmclusters;

		// compute variable parameters in two passes (absolute values in first pass, relative values in second pass)
		size = 0.0;
		for(int i = 0; i < nmclusters; ++i)
		{
			blobs[assigned[i]].adist = sqrt(sqr(bary_x-blobs[assigned[i]].x)+sqr(bary_y-blobs[assigned[i]].y));
			size += blobs[assigned[i]].adist;
			blobs[assigned[i]].aangle = atan2(bary_y-blobs[assigned[i]].y, blobs[assigned[i]].x-bary_x);
// std::cout << "cluster " << i << " bary " << bary_x << "," << bary_y << " ; blob " << blobs[assigned[i]].x << "," << blobs[assigned[i]].y << " ; aangle " << blobs[assigned[i]].aangle*180./M_PI << std::endl;
		}
		size /= nmclusters;

		double prev_angle = blobs[assigned[nmclusters-1]].aangle;
		for(int i = 0; i < nmclusters; ++i)
		{
			blobs[assigned[i]].dist = blobs[assigned[i]].adist / size;
			blobs[assigned[i]].size = blobs[assigned[i]].asize / size;
			blobs[assigned[i]].angle = blobs[assigned[i]].aangle - prev_angle;
			prev_angle = blobs[assigned[i]].aangle;
			constraint_angle(blobs[assigned[i]].angle, blobs[assigned[i]].angle);
		}
	}
	
	///*********************************************************************************************************************************************
	/// match
	///*********************************************************************************************************************************************
	
	void RefinerHueBlobs::match(std::vector<int> & assigned, std::set<int> & remaining, std::vector<Blob> & blobs, Solution & best_solution)
	{
		// TODO [speed] stop distance computation when no hope to improve current min
		// TODO [speed] choose order greedily so that we find best solutions first (at least hue)
		// TODO [precision] hue and angle should be compared modulo...
		if (assigned.size() == nmclusters)
		{
			double bary_x, bary_y, size;
			evalBlobs(assigned, blobs, bary_x, bary_y, size);

			// compute score
			cv::Mat features = cv::Mat(1, model.cols, CV_32FC1);
			for(int i = 0; i < nmclusters; ++i)
			{
				features.at<float>(0,4*i+0) = blobs[assigned[i]].hue;
				features.at<float>(0,4*i+1) = blobs[assigned[i]].dist;
				features.at<float>(0,4*i+2) = blobs[assigned[i]].angle;
				features.at<float>(0,4*i+3) = blobs[assigned[i]].size;
			}
// 			double score = Distance::dist(model, features, model_sig);
// std::cout << "score1 " << score << std::endl;

			double score = 0.0;
			for(int i = 0; i < model.cols; ++i)
			{
				double inc = (model.at<float>(0,i)-features.at<float>(0,i)) / model_sig.at<float>(0,i);
				score += inc*inc;
			}
			score = sqrt(score);

// for(int i = 0; i < nmclusters; ++i) std::cout << assigned[i] << " ";
// std::cout << "(score " << score << ")" << std::endl;

			if (score < best_solution.score)
			{
// std::cout << "new best solution with ";
// for(int i = 0; i < nmclusters; ++i) std::cout << assigned[i] << " ";
// std::cout << "(score " << score << ")" << std::endl;
				best_solution.score = score;
				for(int i = 0; i < nmclusters; ++i) best_solution.blobs[i] = blobs[assigned[i]];
				best_solution.bary_x = bary_x;
				best_solution.bary_y = bary_y;
				best_solution.size = size;
			}

		} else
		// recursive calls
		{
			std::set<int> remaining2 = remaining;
			for(std::set<int>::iterator it = remaining.begin(); it != remaining.end(); ++it)
			{
				assigned.push_back(*it);
				remaining2.erase(*it);
				match(assigned, remaining2, blobs, best_solution);
				remaining2.insert(*it);
				assigned.pop_back();
			}
		}
	}

	///*********************************************************************************************************************************************
	/// computeState
	///*********************************************************************************************************************************************
	
	void RefinerHueBlobs::computeState(Object &object, Solution & best_solution, cv::Point offset)
	{
// for(int i = 0; i < nmclusters; ++i) std::cout << "cluster " << i << " pos " << best_solution.blobs[i].x << "," << best_solution.blobs[i].y << std::endl;
		double size_avg = 0.0, size_weights = 0.0;
		double angle_avg = 0.0, angle_weights = 0.0;
		double first_angle;
		for(int i = 0; i < nmclusters; ++i)
		{
			size_avg += best_solution.blobs[i].asize * ratios.at<float>(0,4*i+0) / ratios_sig.at<float>(0,4*i+0);
			size_weights += 1.0 / ratios_sig.at<float>(0,4*i+0);
			double angle = (best_solution.blobs[i].aangle + ratios.at<float>(0,4*i+1));
			if (i == 0) first_angle = angle; else constraint_angle(angle - first_angle, angle);
// std::cout << "cluster " << i << " size est " << best_solution.blobs[i].asize * ratios.at<float>(0,4*i+0) << " (weight " << 1/ratios_sig.at<float>(0,4*i+0) << ")" << std::endl;
// std::cout << "angle est " << angle << " (weight " << 1/ratios_sig.at<float>(0,4*i+1) << ") from angle " << best_solution.blobs[i].aangle << " and correction " << ratios.at<float>(0,4*i+1) <<  std::endl;
			angle_avg += angle / ratios_sig.at<float>(0,4*i+1);
			angle_weights += 1.0 / ratios_sig.at<float>(0,4*i+1);
		}
// std::cout << "size_avg " << size_avg << " size_weights " << size_weights << " angle_avg " << angle_avg << " angle_weights " << angle_weights << std::endl;
		size_avg /= size_weights;
		angle_avg /= angle_weights;
		constraint_angle(angle_avg, angle_avg);

		// TODO compute center with ratios

		object.set(Object::lpPosU, offset.x + best_solution.bary_x, 4.0);
		object.set(Object::lpPosV, offset.y + best_solution.bary_y, 4.0);
		object.set(Object::lpSize, size_avg, 0.20*size_avg);
		object.set(Object::lpHeading, angle_avg, 2*M_PI/180.);
	}
	

	///*********************************************************************************************************************************************
	/// floodFillExtract
	///*********************************************************************************************************************************************
	// TODO templatize this with image typename, TEST functor, blob and/or update blob stats functor
	//template<class TEST, typename T>
	void RefinerHueBlobs::floodFillExtract(cv::Mat const & image, std::list<Blob> & blobs)
	{
		typedef uchar T;
		cv::Mat pimg = cv::Mat(image.rows, image.cols, CV_8U, cv::Scalar(128)); // progress image: 128 is unknown, 255 is positive, 0 is negative

		blobs.clear();
		std::queue<cv::Point> queue;
		cv::Point nb[4]; // neighbors

		// start by going through the image to find points above threshold
		for(unsigned i = 0; i < image.rows; ++i)
			for(unsigned j = 0; j < image.cols; ++j)
			{
				if (pimg.at<uchar>(i,j) == 128)
					if (image.at<T>(i,j) != 0)
					{
						Blob blob;
						blob.asize = 0.0;
						blob.x = blob.y = 0.0;
						blob.hue = 0.0;
						queue.push(cv::Point(j,i));
						pimg.at<uchar>(i,j) = 255;

						// start flood fill
						while (!queue.empty())
						{
							cv::Point & p = queue.front();

							blob.asize += 1.0;
							blob.x += p.x; blob.y += p.y;
							blob.hue += hue_patch.at<uchar>(p.y,p.x);

							int nk = 0;
							if (p.x-1 >= 0)         { nb[nk].x = p.x-1; nb[nk].y = p.y  ; ++nk; }
							if (p.x+1 < image.cols) { nb[nk].x = p.x+1; nb[nk].y = p.y  ; ++nk; }
							if (p.y-1 >= 0)         { nb[nk].x = p.x  ; nb[nk].y = p.y-1; ++nk; }
							if (p.y+1 < image.rows) { nb[nk].x = p.x  ; nb[nk].y = p.y+1; ++nk; }
							for(int k = 0; k < nk; ++k)
								if (pimg.at<uchar>(nb[k].y,nb[k].x) == 128)
									if (image.at<T>(nb[k].y,nb[k].x) != 0)
									{
										pimg.at<uchar>(nb[k].y,nb[k].x) = 255;
										queue.push(cv::Point(nb[k].x, nb[k].y));
									} else pimg.at<uchar>(nb[k].y,nb[k].x) = 0;

							queue.pop();
						}
						blobs.push_back(blob);
					} else pimg.at<uchar>(i,j) = 0;
			}
	}

	///*********************************************************************************************************************************************
	/// extractClusters
	///*********************************************************************************************************************************************

	void RefinerHueBlobs::extractClusters(cv::Mat & patch, std::vector<Blob> & blobs)
	{
#if DISPLAY
	const int nsteps = 3;
	const int ncolors = nsteps*nsteps*nsteps - nsteps;
	static const uchar steps[nsteps] = {0,128,255};
	uchar colors[ncolors][3];// = {{255,0,0},{0,255,0},{0,0,255},{0,255,255},{255,0,255},{255,255,0},{128,0,0},{0,128,0},{0,0,128},{0,128,128};
	for(int i = 0, c = 0; i < 3; ++i) for(int j = 0; j < 3; ++j) for(int k = 0; k < 3; ++k)
	{
		if (i == j && j == k) continue;
		colors[c][0] = steps[i];
		colors[c][1] = steps[j];
		colors[c][2] = steps[k];
		++c;
	}
#endif

		cv::Mat samples = rgb_order ? createSamples<0,1,2>(patch) : createSamples<2,1,0>(patch);
		cv::Mat labels, centers;
		const int nclusters = 20;
		if (samples.rows < nclusters) { blobs.clear(); return; }

		// kmeans
		cv::kmeans(samples, nclusters, labels, cv::TermCriteria(cv::TermCriteria::COUNT|cv::TermCriteria::EPS, 10, 1e-10), 1, cv::KMEANS_PP_CENTERS, centers);

		// translate pixel list to image clusters
		cv::Mat clusters[nclusters];
		for(int i = 0; i < nclusters; ++i) clusters[i] = cv::Mat(patch.rows, patch.cols, CV_8UC1, cv::Scalar(0));
		for(int i = 0; i < samples.rows; ++i)
			clusters[labels.at<uint>(i)].at<uchar>(samples.at<float>(i,0), samples.at<float>(i,1)) = 255;

#if DISPLAY
	static int nim = 0;
	std::ostringstream oss; oss << std::setw(2) << std::setfill('0') << nim;
	cv::imwrite(oss.str()+std::string("_1patch.png"), patch);
	cv::Mat imblobs = cv::Mat(patch.rows, patch.cols, CV_8UC3, cv::Scalar(0,0,0));
	for(int i = 0; i < nclusters; ++i) for(int y = 0; y < patch.rows; ++y) for(int x = 0; x < patch.cols; ++x)
	if (clusters[i].at<uchar>(y,x) != 0) for(int j = 0; j < 3; ++j) imblobs.at<uchar[3]>(y,x)[j] = colors[i][j];
	cv::imwrite(oss.str()+std::string("_4clusters.png"), imblobs);
#endif
		
		// remove border clusters
		std::vector<uint> iclusters;
		for(int i = 0; i < nclusters; ++i)
		{
			uint nborder = 0;
			for(int j = 0; j < patch.cols  ; ++j) if (clusters[i].at<uchar>(0,j) == 255 || clusters[i].at<uchar>(patch.rows-1,j) == 255) ++nborder;
			for(int j = 1; j < patch.rows-1; ++j) if (clusters[i].at<uchar>(j,0) == 255 || clusters[i].at<uchar>(j,patch.cols-1) == 255) ++nborder;
			if (nborder <= 8) iclusters.push_back(i);
		}


#if DISPLAY
	imblobs = cv::Mat(patch.rows, patch.cols, CV_8UC3, cv::Scalar(0,0,0));
	for(int i = 0; i < iclusters.size(); ++i) for(int y = 0; y < patch.rows; ++y) for(int x = 0; x < patch.cols; ++x)
		if (clusters[iclusters[i]].at<uchar>(y,x) != 0) for(int j = 0; j < 3; ++j) imblobs.at<uchar[3]>(y,x)[j] = colors[i][j];
	cv::imwrite(oss.str()+std::string("_3clusters.png"), imblobs);
#endif

		// remove noise in remaining clusters
		cv::Mat element = cv::Mat(3,3,CV_8UC1,cv::Scalar(0)); for(int i = 0; i < 3; ++i) { element.at<uchar>(1,i) = element.at<uchar>(i,1) = 255; } // cross 3x3
		//cv::Mat element = cv::Mat(3,3,CV_8UC1,cv::Scalar(255)); // square 3x3
		for(int i = 0; i < iclusters.size(); ++i)
		{
			erode(clusters[iclusters[i]], clusters[iclusters[i]], element, cv::Point(-1, -1), 1);
			dilate(clusters[iclusters[i]], clusters[iclusters[i]], element, cv::Point(-1, -1), 1);
//			dilate(clusters[iclusters[i]], clusters[iclusters[i]], element, cv::Point(-1, -1), 1);
//			erode(clusters[iclusters[i]], clusters[iclusters[i]], element, cv::Point(-1, -1), 1);
		}

#if 1 // split
		// TODO split clusters with several blobs... and try to merge intricated ones ?
		blobs.resize(0);
		std::list<Blob> sublobs;
		for(int i = 0; i < iclusters.size(); ++i)
		{
			floodFillExtract(clusters[iclusters[i]], sublobs);
			sublobs.sort();
			unsigned nadd = 0;
			for(std::list<Blob>::iterator it = sublobs.begin(); it != sublobs.end() && nadd < 2; ++it, ++nadd)
			{
				Blob &blob = *it;
				blob.x /= blob.asize;
				blob.y /= blob.asize;
				blob.hue /= blob.asize;
				if (blob.asize >= 6) blobs.push_back(blob);
			}
		}
#else
		// extract blobs
		// TODO use modulo for hue average just like for angle average above
		blobs.resize(0);
		for(int i = 0; i < iclusters.size(); ++i)
		{
			Blob blob;
			blob.asize = 0.0;
			blob.x = blob.y = 0.0;
			blob.hue = 0.0;

			for(int y = 0; y < patch.rows; ++y)
				for(int x = 0; x < patch.cols; ++x)
					if (clusters[iclusters[i]].at<uchar>(y,x) != 0)
						{ blob.asize += 1.0; blob.x += x; blob.y += y; blob.hue += hue_patch.at<uchar>(y,x); }

			if (blob.asize != 0.0)
			{
				blob.x /= blob.asize;
				blob.y /= blob.asize;
				blob.hue /= blob.asize;
				blobs.push_back(blob);
			}
		}
#endif

#if DISPLAY
	imblobs = cv::Mat(patch.rows, patch.cols, CV_8UC3, cv::Scalar(0,0,0));
	for(int i = 0; i < iclusters.size(); ++i) for(int y = 0; y < patch.rows; ++y) for(int x = 0; x < patch.cols; ++x)
		if (clusters[iclusters[i]].at<uchar>(y,x) != 0) for(int j = 0; j < 3; ++j) imblobs.at<uchar[3]>(y,x)[j] = colors[i][j];
	cv::imwrite(oss.str()+std::string("_2clusters.png"), imblobs);
// std::cout << "writing to " << nim << " files" << std::endl;
	++nim;
#endif
		
	}

	
	///*********************************************************************************************************************************************
	/// refine
	///*********************************************************************************************************************************************
	
	bool RefinerHueBlobs::refine(Object &object, cv::Mat const& image, double &distance)
	{
#ifdef CHRONO
	static struct timeval tv1, tv2;
	static double avgt = 0., mint = 1e6;
	static int n = 0;
	gettimeofday(&tv1,NULL);
#endif

		cv::Point offset;
		cv::Mat patch = extractPatch(object, image, offset);

		std::vector<Blob> blobs;
		extractClusters(patch, blobs);
// std::cout << "found " << blobs.size() << " clusters:" << std::endl;
// for(int i = 0; i < blobs.size(); ++i) std::cout << i << " " << blobs[i].x << "," << blobs[i].y << std::endl;

		// compute
		Solution best_solution;
		best_solution.score = 1e10;
		if (blobs.size() >= nmclusters)
		{
			best_solution.blobs.resize(nmclusters);
			std::vector<int> assigned;
			std::set<int> clusters;
			for(int i = 0; i < blobs.size(); ++i) clusters.insert(i);
			match(assigned, clusters, blobs, best_solution);
		}
		distance = best_solution.score;
		object.score = exp(-best_solution.score);

#ifdef CHRONO
	gettimeofday(&tv2,NULL);
	double t = (tv2.tv_sec-tv1.tv_sec) + (tv2.tv_usec-tv1.tv_usec)*1e-6;
	avgt += t; if (t < mint) mint = t; ++n;
	std::cout << "[ObjectRefiner] Processed image in " << t << " seconds (average " << avgt/n << " , min " << mint << ")" << std::endl;
#endif

// std::cout << "best combi has score " << distance << " threshold " << threshold << std::endl;
		// test
		if (best_solution.score < threshold)
		{
			computeState(object, best_solution, offset);
			return true;
		} else return false;
		
	}




}

#endif
